﻿{$WARNINGS OFF}
//{$WARN SYMBOL_PLATFORM OFF}
//{$WARN UNIT_PLATFORM OFF}
//=============================================================================
//  他のアプリのウィンドウにショートカットキー送信するための手続き類
//
//-----------------------------------------------------------------------------
//
//  【履歴】
//
//  2021年08月04日 ～
//
//-----------------------------------------------------------------------------
//
//  【動作確認環境】
//
//  Windows [Version 10.0.19042] Pro 64 bit + Delphi XE5(UP2) Pro VCL-32
//  添付の plvkEnum.pas を使用

//
//  Presented by Mr.XRAY
//  http://mrxray.on.coocan.jp/
//=============================================================================
unit plSendShortCut;

interface

uses
  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Classes, Vcl.Dialogs;

  function SendShortCutKeyEx(hWindow: HWND; AKeyText: string; ACount: Integer=1):
    string; overload;
  function SendShortCutKeyEx(AKeyText: string;
    ACount: Integer=1): string; overload;
  function SendShortCutKeyEx(hWindow: HWND; AShortCut: TShortCut): string; overload;
  function SendShortCutKeyEx(AShortCut: TShortCut): string; overload;

  procedure SendKeysASCII(hWindow: HWND; AKeys: string); overload;
  procedure SendKeysASCII(AKeys: string); overload;
  procedure SendInputText(hWindow: HWND;　AText: string); overload;
  procedure SendInputText(AText: string); overload;
  procedure SendShortCutKey(hWindow: HWND; AKeyChars: string;
    ACount: Integer=1); overload;
  procedure SendShortCutKey(AKeyChars: string; ACount: Integer=1); overload;
  function GetShorCutChars(AShorCutText: string): string;
  function AsciiToVirtualKey(AChar: Char): Byte; overload;
  function AsciiToVirtualKey(AKeyByte: Byte): Byte; overload;

implementation

uses
  System.StrUtils,
  plvkEnum;

//-----------------------------------------------------------------------------
//  ショートカットキー送信用関数
//
//  引数には以下のような文字列でショートカットキーを指定する
//    Ctrl + C
//    Ctrl + Shift + A
//    Alt + F4 
//  VK_SNAPSHOT, vkSnapShot 等の文字列も使用可能
//
//  Shift + Right のショートカットで，Right キーだけを繰り返し操作する場合
//  ACount にその回数を指定する 
//-----------------------------------------------------------------------------
function SendShortCutKeyEx(hWindow: HWND; AKeyText: string;
  ACount: Integer=1): string;
begin
  Result:= GetShorCutChars(AKeyText);
  SendShortCutKey(hWindow, Result, ACount);
end;

//-----------------------------------------------------------------------------
//  ショートカットキー送信用関数
//  アクティブなウィンドウ用
//-----------------------------------------------------------------------------
function SendShortCutKeyEx(AKeyText: string; ACount: Integer=1): string;
begin
  Result:= GetShorCutChars(AKeyText);
  SendShortCutKey(Result, ACount);
end;

//-----------------------------------------------------------------------------
//  ショートカットキー送信用関数
//  引数が TShortCut 型の場合
//
//  TShortCut 型の値は TMenuItem.ShortCut プロパティで取得できる
//-----------------------------------------------------------------------------
function SendShortCutKeyEx(hWindow: HWND; AShortCut: TShortCut): string;
begin
  if hWindow = 0 then Exit;

  // hWindowは送信先ウィドウのウィンドウハンドル
  SetForegroundWindow(hWindow);
  SendShortCutKeyEx(AShortCut);
end;

//-----------------------------------------------------------------------------
//  ショートカットキー送信用関数
//  引数が TShortCut 型の場合
//  アクティブなウィンドウ用
//-----------------------------------------------------------------------------
function SendShortCutKeyEx(AShortCut: TShortCut): string;
var
  LShift : string;
begin
  // Vcl.Menus.ShortCutToText の結果は使用できない文字があるので使用しない
  LShift := '';
  if AShortCut and $2000 <> 0 then LShift := LShift + Char(VK_SHIFT);
  if AShortCut and $4000 <> 0 then LShift := LShift + Char(VK_CONTROL);
  if AShortCut and $8000 <> 0 then LShift := LShift + Char(VK_MENU);

  Result := LShift + Char(Byte(AShortCut));
  SendShortCutKey(Result);
end;

//-----------------------------------------------------------------------------
//  指定ウィンドウに文字として送信
//  送信可能なのは ASCII 文字の範囲
//
//  受信側では送信した仮想キーコードがそのまま取得できる
//  英大文字は Shift キー ($10) に続いて該当の仮想キーコード
//-----------------------------------------------------------------------------
procedure SendKeysASCII(hWindow: HWND; AKeys: string);
begin
  if hWindow = 0 then Exit;

  // hWindowは送信先ウィドウのウィンドウハンドル
  SetForegroundWindow(hWindow);
  SendKeysASCII(AKeys);
end;

//-----------------------------------------------------------------------------
//  アクティブなウィンドウに文字として送信
//-----------------------------------------------------------------------------
procedure SendKeysASCII(AKeys: string);
var
  LIndex    : Integer;
  LKeyWord  : WORD;
  LKeyByte  : Byte;
  LKeyState : TKeyboardState;
  LIsShift  : Boolean;
begin
  // 改行文字変更
  AKeys :=  AdjustLineBreaks(AKeys, tlbsLF);

  for LIndex := 1 to Length(AKeys) do begin
    // ASCII 文字を仮想キーコードの値に変換
    LKeyWord := VkKeyScanEx(Char(AKeys[LIndex]), GetKeyboardLayout(0));

    LIsShift := (LKeyWord and $100) <> 0;
    if Lo(LKeyWord) in [$41..$5A] then begin
      // 数字キーは Caps Lock されていたら Shift キー押下なしにする
      GetKeyboardState(LKeyState);
      if LKeyState[VK_CAPITAL] = 1 then LIsShift := not LIsShift;
    end;

    if LIsShift then begin
      keybd_event(VK_SHIFT, MapVirtualKey(VK_SHIFT, 0), 0, 0);
    end;
    LKeyByte := Lo(LKeyWord);

    keybd_event(LKeyByte, MapVirtualKey(LKeyByte, 0),0, 0);
    keybd_event(LKeyByte, MapVirtualKey(LKeyByte, 0), KEYEVENTF_KEYUP, 0);

    if LIsShift then begin
      keybd_event(VK_SHIFT, MapVirtualKey(VK_SHIFT, 0), KEYEVENTF_KEYUP, 0);
    end;
  end;
  Sleep(10);
end;

//-----------------------------------------------------------------------------
//  SendInput 関数を使用した文字列 (漢字を含む) の送信
//
//  TInput.ki.dwFlags に KEYEVENTF_UNICODE を含め
//  TInput.ki.wVk := 0 にすると文字の送信になる
//
//  受信側で受信できるのは文字数分の $E7 (キーボード以外からの入力手段)
//-----------------------------------------------------------------------------
procedure SendInputText(hWindow: HWND;　AText: string);
begin
  if hWindow = 0 then Exit;

  // hWindowは送信先ウィドウのウィンドウハンドル
  SetForegroundWindow(hWindow);
  SendInputText(AText);
end;

//-----------------------------------------------------------------------------
//  SendInput 関数を使用した文字列 (漢字を含む) の送信
// アクティブなウィンドウ用
//-----------------------------------------------------------------------------
procedure SendInputText(AText: string);
var
  LInput : array [0..1] of TInput;
  LIndex : Integer;
begin
  // 改行文字変換
  AText := AdjustLineBreaks(AText, tlbsLF);

  for LIndex := 1 to Length(AText) do begin
    // キーを押す設定
    LInput[0].Itype          := INPUT_KEYBOARD;
    LInput[0].ki.wVk         := 0;
    LInput[0].ki.wScan       := Ord(AText[lIndex]);
    LInput[0].ki.dwFlags     := KEYEVENTF_UNICODE;
    LInput[0].ki.time        := 0;
    LInput[0].ki.dwExtraInfo := 0;

    // キーを離す設定
    LInput[1].Itype          := INPUT_KEYBOARD;
    LInput[1].ki.wVk         := 0;
    LInput[1].ki.wScan       := Ord(AText[lIndex]);
    LInput[1].ki.dwFlags     := KEYEVENTF_KEYUP or KEYEVENTF_UNICODE;
    LInput[1].ki.time        := 0;
    LInput[1].ki.dwExtraInfo := 0;

    SendInput(2, LInput[0], SizeOf(TInput));
  end;
  Sleep(10);
end;

//-----------------------------------------------------------------------------
//  指定ウィンドウにショートカットキーを送信
//
//  ACmdWnd   : 対象のウィンドウハンドル
//  AKeyChars : 仮想キーコードの文字列．GetShorCutChara 関数で作成できる
//              VK_CONTROL + C (Ctrl + C) の場合は #$11#$43 を渡す
//              これはChr(VK_CONTROL) + Chr('C') と同じ
//  ACount  　: ショートカッキー操作の繰り返し回数
//              Ctrl + Right 等の Right キーだけを複数回繰り返し実行したい場合に指定
//              繰り返すのは最後のキーのみ
//
//  keybd_event 関数では，拡張キーの場合，KEYEVENTF_EXTENDEDKEY が必要
//  拡張キーは，主にテンキーの左側にある矢印キー等
//  [Windows] キーも拡張キー
//-----------------------------------------------------------------------------
procedure SendShortCutKey(hWindow: HWND; AKeyChars: string; ACount: Integer=1);
begin
  if hWindow = 0 then Exit;

  // hWindowは送信先ウィドウのウィンドウハンドル
  SetForegroundWindow(hWindow);
  SendShortCutKey(AKeyChars, ACount);
end;

//-----------------------------------------------------------------------------
//  アクティブなウィンドウにショートカットキーを送信
//-----------------------------------------------------------------------------
procedure SendShortCutKey(AKeyChars: string; ACount: Integer=1);

　//---------------------------------------------------------------------------
  procedure Keyevent(AByte: Byte; AFlag: Cardinal);
  const
    // 拡張キーの判定用
    LExtendKys = [$21..$2F, $5B..$5F, $A6..$B7, $90];
  var
    LKeyFlag : Cardinal;
  begin
    if AByte in LExtendKys then begin
      LKeyFlag := AFlag or KEYEVENTF_EXTENDEDKEY;
    end else begin
      LKeyFlag := AFlag;
    end;
    keybd_event(AByte, MapVirtualKey(AByte, 0), LKeyFlag, 0);
  end;
　//---------------------------------------------------------------------------

var
  LIndex    : Integer;
  LCount    : Integer;
  LKeyByte  : Byte;
begin
  if AKeyChars = '' then Exit;

  // キーを押したままの状態にする
  for LIndex := 1 to Length(AKeyChars) - 1 do begin
    LKeyByte := Ord(AKeyChars[LIndex]);
    Keyevent(LKeyByte, 0);
  end;

  // 最後のキーの繰り返し操作
  // 最低 1 回は実行
  LKeyByte := Ord(AKeyChars[Length(AKeyChars)]);
  for LCount := 1 to ACount do begin
    Keyevent(LKeyByte, 0);
    Keyevent(LKeyByte, KEYEVENTF_KEYUP);
  end;

  // キーを順番に離す操作
  for LIndex := Length(AKeyChars) - 1 downto 1 do begin
    LKeyByte := Ord(AKeyChars[LIndex]);
    Keyevent(LKeyByte, KEYEVENTF_KEYUP);
  end;
  // 少しタイムラグが必要 (キーバッファの処理か ?)
  Sleep(10);
end;

//-----------------------------------------------------------------------------
//  'Ctrl + C' 等の文字列を仮想キーコードの文字列に変換
//   + の両端に半角の空白があってもよい
//  文字を [+] のように [] で括ることも可能
//   'Ctrl+C' (または 'VK_CONTROL + C') の時は #$11#$43 を戻り値として返す
//
//  [Fn] キーには仮想キーコードはない ([Fn] + XX で特定の値が割り当てられている
//
//  uses に System.StrUtils が必要
//  uses に plvkEnum (添付のユニット) が必要
//-----------------------------------------------------------------------------
function GetShorCutChars(AShorCutText: string): string;
var
  LStrList : TStringList;
  LArgText : string;
  LvkText  : string;
  LResStr  : string;
  LByte    : Byte;
  LIndex   : Integer;
begin
  Result := '';
  if AShorCutText = '' then Exit;
  if Trim(AShorCutText)[1]  = '+' then Exit;

  AShorCutText := ReplaceText(AShorCutText, ' ', '');
  AShorCutText := ReplaceText(AShorCutText, '[+]', 'VK_ADD');
  AShorCutText := ReplaceText(AShorCutText, '++',  '+VK_ADD');

  LResStr := '';
  LStrList := TStringList.Create;
  try
    LStrList.Delimiter       := '+';
    LStrList.StrictDelimiter := True;
    LStrList.DelimitedText   := AShorCutText;

    for LIndex := 0 to LStrList.Count - 1 do begin
      LArgText := LStrList[LIndex];
      if (LeftStr(LArgText, 1) = '[') and (RightStr(LArgText, 1) = ']') then begin
        LArgText := Copy(LArgText, 2, Length(LArgText) -2);
      end;
      LvkText := UpperCase(LArgText);

      // 仮想キーコード名として使用可能な省略形式の文字列の判定
      // VK_XXX の VK_ も省略可能
      if LvkText = '←'          then LvkText := 'VK_LEFT';
      if LvkText = '↑'          then LvkText := 'VK_UP';
      if LvkText = '→'          then LvkText := 'VK_RIGHT';
      if LvkText = '↓'          then LvkText := 'VK_DOWN';
      if LvkText = 'ENTER'       then LvkText := 'VK_RETURN';
      if LvkText = 'CTRL'        then LvkText := 'VK_CONTROL';
      if LvkText = 'ALT'         then LvkText := 'VK_MENU';
      if LvkText = 'CAPSLOCK'    then LvkText := 'VK_CAPITAL';
      if LvkText = 'ESC'         then LvkText := 'VK_ESCAPE';
      if LvkText = 'PRTSC'       then LvkText := 'VK_SNAPSHOT';
      if LvkText = 'PRTSCN'      then LvkText := 'VK_SNAPSHOT';
      if LvkText = 'PRINTSCREEN' then LvkText := 'VK_SNAPSHOT';
      if LvkText = 'INS'         then LvkText := 'VK_INSERT';
      if LvkText = 'DEL'         then LvkText := 'VK_DELETE';
      if LvkText = 'PAGEUP'      then LvkText := 'VK_PRIOR';
      if LvkText = 'PAGEDOWN'    then LvkText := 'VK_NEXT';
      if LvkText = 'WINDOWS'     then LvkText := 'VK_LWIN';
      if LvkText = 'WIN'         then LvkText := 'VK_LWIN';

      if (Length(LArgText) > 1) and (LArgText[1] <> '$') then begin
        if LeftStr(UpperCase(LvkText), 2) <> 'VK' then LvkText := 'vk' + LvkText;
      end;

      LByte := vkTextToVirtualKey(LvkText);
      if LByte = $FF then Break;
      LResStr := LResStr + Char(LByte);
    end;
    Result := LResStr;
  finally
    FreeAndNil(LStrList);
  end;
end;

//-----------------------------------------------------------------------------
//  ASCII 文字を仮想キーーコードの値に変換する関数
//  文字 = と - を変換すると共に $BD になる
//
//  $20..$2F には VK_SPACE..VK_HELP までが割り当てられている　
//           ASCII 文字は !"#$%&'()*+,-./ ($20 は [Space])
//  $3A..$40 は未定義
//           ASCII 文字は :;<=>?@
//  $5B..$60 には VK_LWIN..VK_NUMPAD0 までが割り当てられている
//           ASCII 文字は [\]^_`
//  $7B..$7E には VK_F12..VK_F15 までが割り当てられている
//           ASCII 文字は {|}~
//-----------------------------------------------------------------------------
function AsciiToVirtualKey(AChar: Char): Byte;
var
  LKeyByte : Byte;
  LLayout  : HKL;
begin
  LKeyByte := Byte(AChar);
  Result := LKeyByte;
  if LKeyByte in [$20..$2F, $3A..$40, $5B..$60, $7B..$7E] then begin
    LLayout :=  GetKeyboardLayout(0);
    Result := Lo(VkKeyScanEx(AChar, LLayout));
  end;
end;

//-----------------------------------------------------------------------------
//  ASCII 文字コードを仮想キーコードの値に変換する関数
//-----------------------------------------------------------------------------
function AsciiToVirtualKey(AKeyByte: Byte): Byte;
begin
  Result := AsciiToVirtualKey(Char(AKeyByte));
end;

end.
